/*
 input/Zapper.c

 * MacFCEU 0.9d6
 *
 * Copyright notice for this file:
 *  zapper.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "share.h"
#include "../video.h"
#include "../palette.h"
#include "../git.h"
#include "Zapper.h"

/* ---- constants */

#define RANGE_TOLERANCE		1 /* was 4 */

/* ---- globals */

ZAPPER		InputZapperData[3];

/* ---- code */

void FP_FASTAPASS(3) ZapperFrapper(int w,
	uint8 *bg,
	uint8 *spr,
	uint32 linets,
	int final)
{
	ZAPPER		*zp = InputZapperData+w;
	int		xs,xe;
	int		zx,zy;

	if (!bg) // New line, so reset stuff.
	{
		zp->zappo = 0;
		return;
	}
	xs = zp->zappo;
	xe = final;

	zx = zp->mzx;
	zy = zp->mzy;

	if (xe>256)
		xe = 256;

	if (scanline>=zy-RANGE_TOLERANCE && scanline<=zy+RANGE_TOLERANCE)
	{
		while(xs<xe)
		{
			uint32		sum;
			uint8		px,oa;

			if(xs<=(zx+RANGE_TOLERANCE) && xs>=(zx-RANGE_TOLERANCE))
			{
				px = bg[xs];
				if (spr)
					if ((oa=spr[xs])!=0 && (uint8)((oa|px)&PLB_B))
						px = oa;
				px&= 63;
	
				sum = palo[px].r+palo[px].g+palo[px].b;
				if (sum>=100*3)
				{
					uint64		t;
	
					//InputZapperData[w].zaphit=((uint64)linets+(xs+16)*(PAL?15:16))/48+timestampbase; 
					(void)mulu64u32(t,xs+16,PAL ? CPU_SUBCYC*5/16:CPU_SUBCYC*4/12);
					(void)uint64addu32(t,linets);
					(void)uint64divu32(t,CPU_SUBCYC);
					(void)uint64add(t,timestampbase);
					zp->zaphit = t;
					goto endo;
				}
			}   
			xs++;
		}
	}
endo:
	zp->zappo=final;
}      

static INLINE int CheckColor(int w)
{
	uint64		t, u;

	FCEUPPU_LineUpdate();
/*if((ZD[w].zaphit+100)>=(timestampbase+timestamp) && !(ZD[w].mzb&2)) return(0); */
	t = InputZapperData[w].zaphit;
	(void)uint64addu32(t,100);
	u = timestampbase;
	(void)uint64addu32(u,timestamp);
	return int64ge(t,u) && !(InputZapperData[w].mzb&2) ? 0:1; //2011.02.05
}

static uint8 FP_FASTAPASS(1) ReadZapperVS(int w)
{
	uint8	ret = 0;

	if (InputZapperData[w].zap_readbit==4)
		ret = 1;

	if (InputZapperData[w].zap_readbit==7)
		if (InputZapperData[w].bogo)
			ret|= 1;
	if (InputZapperData[w].zap_readbit==6)
		ret|= CheckColor(w);
	if (!fceuindbg)
		++InputZapperData[w].zap_readbit; 
	return ret;
}

static void FP_FASTAPASS(1) StrobeZapperVS(int w)
{
	InputZapperData[w].zap_readbit = 0;
}

static uint8 FP_FASTAPASS(1) ReadZapper(int w)
{
	uint8		ret = 0;

	if (InputZapperData[w].bogo)
		ret|= 0x10;
	ret|= CheckColor(w)<<3; //if...|=8
	return ret;
}

static void FASTAPASS(3) DrawZapper(int w,
	uint8 *buf,
	int arg)
{
	if (arg)
		FCEU_DrawGunSight(buf,InputZapperData[w].mzx,InputZapperData[w].mzy);
}

static void FP_FASTAPASS(3) UpdateZapper(int w,
	void *data,
	int arg)
{
	uint32		*ptr = (uint32*)data;
(void)arg;
//FCEU_DispMessage("%3d:%3d",ZD[w].mzx,ZD[w].mzy);
#if 0
	InputZapperData[w].bogo = (ptr[2]&3)!=0;
#else
	if (InputZapperData[w].bogo)
		--InputZapperData[w].bogo;
	if (ptr[2]&3 && (InputZapperData[w].mzb&3)==0)
		InputZapperData[w].bogo = 5;
#endif
	InputZapperData[w].mzx = ptr[0];
	InputZapperData[w].mzy = ptr[1];
	InputZapperData[w].mzb = ptr[2];
}

static INPUTC ZAPC =
	{ReadZapper,0,0,UpdateZapper,ZapperFrapper,DrawZapper};
static INPUTC ZAPVSC =
	{ReadZapperVS,0,StrobeZapperVS,UpdateZapper,ZapperFrapper,DrawZapper};

INPUTC *FCEU_InitZapper(int w)
{
	memset(&InputZapperData[w],0,sizeof(ZAPPER));
	if (FCEUGameInfo->type==GIT_VSUNI)
		return &ZAPVSC;
	else
		return &ZAPC;
}
